#include <stdio.h>
#include <stdlib.h>

#include <xc.h>
#include <pic18.h>


/****************** Eines de Disseny - 2021 *********************/

/*****************************************************************
 *                 Bits de configuracio del PIC                  *
 *****************************************************************/

#pragma config OSC = HSPLL      // Oscillator Selection bits (HS oscillator)
#pragma config OSCS = ON        // Oscillator System Clock Switch Enable bit

// CONFIG2L
#pragma config PWRT = OFF       // Power-up Timer Enable bit (PWRT disabled)
#pragma config BOR = ON         // Brown-out Reset Enable bit (Brown-out Reset enabled)
#pragma config BORV = 25        // Brown-out Reset Voltage bits (VBOR set to 2.5V)

// CONFIG2H
#pragma config WDT = OFF        // Watchdog Timer Enable bit (WDT disabled (control is placed on the SWDTEN bit))
#pragma config WDTPS = 128      // Watchdog Timer Postscale Select bits (1:128)

// CONFIG4L
#pragma config STVR = ON        // Stack Full/Underflow Reset Enable bit (Stack Full/Underflow will cause Reset)
#pragma config LVP = OFF        // Low-Voltage ICSP Enable bit (Low-Voltage ICSP enabled)

// CONFIG5L
#pragma config CP0 = OFF        // Code Protection bit (Block 0 (000200-001FFFh) not code protected)
#pragma config CP1 = OFF        // Code Protection bit (Block 1 (002000-003FFFh) not code protected)
#pragma config CP2 = OFF        // Code Protection bit (Block 2 (004000-005FFFh) not code protected)
#pragma config CP3 = OFF        // Code Protection bit (Block 3 (006000-007FFFh) not code protected)

// CONFIG5H
#pragma config CPB = OFF        // Boot Block Code Protection bit (Boot Block (000000-0001FFh) not code protected)
#pragma config CPD = OFF        // Data EEPROM Code Protection bit (Data EEPROM not code protected)

// CONFIG6L
#pragma config WRT0 = OFF       // Write Protection bit (Block 0 (000200-001FFFh) not write protected)
#pragma config WRT1 = OFF       // Write Protection bit (Block 1 (002000-003FFFh) not write protected)
#pragma config WRT2 = OFF       // Write Protection bit (Block 2 (004000-005FFFh) not write protected)
#pragma config WRT3 = OFF       // Write Protection bit (Block 3 (006000-007FFFh) not write protected)

// CONFIG6H
#pragma config WRTC = OFF       // Configuration Register Write Protection bit (Configuration registers (300000-3000FFh) not write protected)
#pragma config WRTB = OFF       // Boot Block Write Protection bit (Boot Block (000000-0001FFh) not write protected)
#pragma config WRTD = OFF       // Data EEPROM Write Protection bit (Data EEPROM not write protected)

// CONFIG7L
#pragma config EBTR0 = OFF      // Table Read Protection bit (Block 0 (000200-001FFFh) not protected from Table Reads executed in other blocks)
#pragma config EBTR1 = OFF      // Table Read Protection bit (Block 1 (002000-003FFFh) not protected from Table Reads executed in other blocks)
#pragma config EBTR2 = OFF      // Table Read Protection bit (Block 2 (004000-005FFFh) not protected from Table Reads executed in other blocks)
#pragma config EBTR3 = OFF      // Table Read Protection bit (Block 3 (006000-007FFFh) not protected from Table Reads executed in other blocks)

// CONFIG7H
#pragma config EBTRB = OFF      // Boot Block Table Read Protection bit (Boot Block (000000-0001FFh) not protected from Table Reads executed in other blocks)

/*****************************************************************
 *                 Definicio de parametres utils                 *
 *****************************************************************/
#define F_CPU 4000000 // oscillator frequency

#define INPUT  1
#define OUTPUT 0
    
#define STATE_LOW  0
#define STATE_HIGH 1

#define BIT0  0x01
#define BIT1  0x02
#define BIT2  0x04
#define BIT3  0x08
#define BIT4  0x10
#define BIT5  0x20
#define BIT6  0x40

// Input pins
#define M1_BWD_SW   PORTAbits.RA1
#define M1_FWD_SW   PORTAbits.RA2
#define M2_BWD_SW   PORTAbits.RA3
#define M2_FWD_SW   PORTAbits.RA4
#define LED_SW      PORTCbits.RC3
    
// Output pins
#define M1_EN       LATDbits.LATD6
#define M1_BWD      LATCbits.LATC4
#define M1_FWD      LATCbits.LATC5
#define M2_EN       LATDbits.LATD7
#define M2_BWD      LATBbits.LATB0
#define M2_FWD      LATBbits.LATB1
#define LED_enabled LATDbits.LATD4
    
// UART baud rate
#define BAUDRATE 9600

// UART codes
#define M1_BWD_UART     1
#define M1_FWD_UART     2
#define M2_BWD_UART     3
#define M2_FWD_UART     4
#define LED_SW_UART     5
#define NOP_UART        85

char T0_OFint_f;   // timer0 overflow interrupt flag
char led_state;    // previous led state (for edge detection)

/*****************************************************************
 *                    Configuracio dels pins                     *
 *****************************************************************/
void init_GPIOs(void)
{
    // digital inputs only
    ADCON0bits.ADON = 0;
    ADCON1bits.PCFG3 = 0;
    ADCON1bits.PCFG2 = 1;
    ADCON1bits.PCFG1 = 1;
    // internal pullup for port B
    INTCON2bits.RBPU = 0;
    PORTB = 0x00;
    LATB = 0x00;
    
    // inputs 
    TRISAbits.RA1 = INPUT; //M1_BWD_SW
    TRISAbits.RA2 = INPUT; //M1_FWD_SW
    TRISAbits.RA3 = INPUT; //M2_BWD_SW
    TRISAbits.RA4 = INPUT; //M2_FWD_SW
    TRISCbits.RC3 = INPUT; //LED_SW
  
    // outputs
    TRISDbits.RD6 = OUTPUT; //M1_ENABLE
    TRISCbits.RC4 = OUTPUT; //M1_BWD
    TRISCbits.RC5 = OUTPUT; //M1_FWD
    TRISDbits.RD7 = OUTPUT; //M2_ENABLE
    TRISBbits.RB0 = OUTPUT; //M2_BWD
    TRISBbits.RB1 = OUTPUT; //M2_FWD
    TRISDbits.RD4 = OUTPUT; //PWM_LED
    
    // initial output states
    M1_EN   = STATE_LOW;
    M1_BWD  = STATE_LOW;
    M1_FWD  = STATE_LOW;
    
    M2_EN   = STATE_LOW;
    M2_BWD  = STATE_LOW;
    M2_FWD  = STATE_LOW;
    
    LED_enabled = STATE_LOW;
}

/****************************************************************
 *               Configuracio del Timer pel delay               *
 ****************************************************************/
void init_TIMER0(void)
{
    T0CONbits.TMR0ON = 0; // stop timer
    T0CONbits.T08BIT = 0; // 16 bit counter
    T0CONbits.T0CS   = 0; // CLKO source
    T0CONbits.T0SE   = 0; // low to high transition
    T0CONbits.PSA    = 0; // preescaler
    
    // prescaler value = 1:256
    T0CONbits.T0PS2  = 1;
    T0CONbits.T0PS1  = 1;
    T0CONbits.T0PS0  = 1;
    
    // enable Timer0 overflow interrupt
    INTCONbits.TMR0IE = 1;
}

/****************************************************************
 *                 Funcio per generar un Delay                  *
 ****************************************************************/
void T0_Delay(unsigned int Ncycles)
{
    // cycles to count
    unsigned int cNcycles = 0xFFFF - Ncycles; // complementary
    TMR0H = (cNcycles & 0xFF00) >> 8; // 16 bit counter higher byte
    TMR0L = (cNcycles & 0x00FF)     ; // 16 bit counter lower byte
    
    T0CONbits.TMR0ON = 1; // start Timer0
}

/********************************************************************
 *                    Inicialitzacio de la UART                     *
 *       Cal indicar el baud rate (velocitat de transmissio)        *
 ********************************************************************/
void init_UART(unsigned long baud_rate)
{
    TRISCbits.RC6=0;     //TX
    TRISCbits.RC7=1;     //RX
    
    int bps = ((F_CPU / baud_rate) >> 2) - 1; // sets baud rate
    SPBRG = bps;         // store baud rate value
    
    TXSTAbits.CSRC = 0;  // don't care for asynchronous mode
    TXSTAbits.TX9  = 0;  // selects 8-bit data transmission
    TXSTAbits.TXEN = 1;  // enable Data tranmission on RC6 pin
    TXSTAbits.SYNC = 0;  // selects Asynchronous Serial Communication Mode
    TXSTAbits.BRGH = 1;  // high speed 
    RCSTAbits.SPEN = 1;  // serial port enabled (RX & TX as serial port pins)
    RCSTAbits.CREN = 1;  // continuius recive until enable bit CREN is cleared
}

/********************************************************************
 *                 Funcio d'escriptura de la UART                   *
 ********************************************************************/
void UART_write(char data)
{
    TXREG = data;
    while(TRMT == 0);
}

char check_in(void)
{
    char var_check;
    
    if (M1_BWD_SW == STATE_HIGH)
        var_check = M1_BWD_UART;
    else if (M1_FWD_SW == STATE_HIGH)
        var_check = M1_FWD_UART;
    else if (M2_BWD_SW == STATE_HIGH)
        var_check = M2_BWD_UART;
    else if (M2_FWD_SW == STATE_HIGH)
        var_check = M2_FWD_UART;
    else if (LED_SW == STATE_HIGH)
        var_check = LED_SW_UART;
    else
        var_check = NOP_UART;
    
    return var_check;
}

/********************************************************************
 *           Commutacio dels outputs en funcio dels inputs          *
 *******************************************************************/
void set_out()
{       
    // motor 1 control
    if(M1_FWD_SW != M1_BWD_SW){
        M1_EN  = STATE_HIGH;
        M1_FWD = M1_FWD_SW;
        M1_BWD = M1_BWD_SW;
    }
    else
        M1_EN  = STATE_LOW;
    
    // motor 2 control
    if(M2_FWD_SW != M2_BWD_SW){
        M2_EN  = STATE_HIGH;
        M2_FWD = M2_FWD_SW;
        M2_BWD = M2_BWD_SW;
    }
    else
        M2_EN  = STATE_LOW;
    
    // led control
    if ((led_state == 0) && (LED_SW == STATE_HIGH))
        LED_enabled ^= 1;       
}

/********************************************************************
 *          Funcio de control de les interrupcions globals          *
 ********************************************************************/
void global_int(char state)
{
    INTCONbits.GIE = state; // global interrupt enable bit
}

/********************************************************************
 *        Funcio main del nostre programa (el que s'executa)        *
 ********************************************************************/
int main(int argc, char** argv) 
{
    /*Inicialitzacio dels moduls necessaris*/
    init_GPIOs();
    init_TIMER0();
    
    init_UART(BAUDRATE);
    char to_send;
    
    /*Variables d'estat per la conmutacio*/
    T0_OFint_f = 1;
    led_state = 0;
    
    /*Activacio interrupcions globals*/
    global_int(1);

    while(1)
    {        
        if(T0_OFint_f)
        {
            to_send = check_in();   // check input pins status
            
            UART_write(to_send);    // send input pins status via uart
            
            set_out();              // set output pins according to input pins
      
            led_state = LED_SW;     // save current led state 
            
            T0_OFint_f = 0;         // clear timer flag            
            T0_Delay(2000);         // wait Ncycles until next iteration
        }
    }    
    return (EXIT_SUCCESS);
}

/********************************************************************
*                       Rutina d'interrupcio                        *
********************************************************************/
void __interrupt() ISR()
{
    // Timer0 overflow interrupt
    if(INTCONbits.TMR0IF){
        T0CONbits.TMR0ON = 0;  // stop Timer0
        T0_OFint_f = 1;
        INTCONbits.TMR0IF = 0; // clear Timer0 overflow interrupt flag
    }
}